package dev.pfaff.altargui.render.objects;

import dev.pfaff.altargui.CanvasImpl;
import dev.pfaff.altargui.render.BufferOverflowException;
import dev.pfaff.altargui.render.LayeredRenderBuffer;
import dev.pfaff.altargui.render.RenderObject;
import dev.pfaff.altargui.render.ViaIntermediateBuffer;

import java.lang.foreign.Arena;
import java.lang.foreign.MemoryLayout;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.StructLayout;
import java.lang.invoke.VarHandle;
import java.nio.ByteOrder;
import java.util.function.Consumer;

import static dev.pfaff.altargui.shader.Programs.textRenderLayer;
import static dev.pfaff.altargui.util.MethodHandleUtil.withoutBase;
import static java.lang.foreign.MemoryLayout.PathElement.groupElement;
import static java.lang.foreign.MemoryLayout.structLayout;
import static java.lang.foreign.ValueLayout.*;

/**
 * A multi-character text render object with individual styling.
 */
public final class TextRenderObject extends RenderObject {
	public static final StructLayout TEXT_CHAR_LAYOUT = structLayout(
			structLayout(JAVA_FLOAT.withName("x"), JAVA_FLOAT.withName("y")).withName("pos"),
			JAVA_FLOAT.withName("size"),
			// bvec4 must be laid out as RGBA. To write an RGBA int and have it be in the correct order, we must write
			// it in BIG_ENDIAN order.
			JAVA_INT.withName("colorRgba").withOrder(ByteOrder.BIG_ENDIAN),
			JAVA_BYTE.withName("char"),
			JAVA_BYTE.withName("flags"),
			JAVA_BYTE.withName("UBase"),
			JAVA_BYTE.withName("VBase")
	);
	public static final VarHandle VH_posX = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
					groupElement("pos"), groupElement("x"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_posY = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
					groupElement("pos"), groupElement("y"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_size = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
					groupElement("size"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_color = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
					groupElement("colorRgba"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_char = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
					groupElement("char"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_flags = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
					groupElement("flags"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_uBase = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
					groupElement("UBase"))
			.withInvokeExactBehavior());
	public static final VarHandle VH_vBase = withoutBase(TEXT_CHAR_LAYOUT.arrayElementVarHandle(
					groupElement("VBase"))
			.withInvokeExactBehavior());

	private float x, y;
	private MemorySegment ms;
	private int length;

	public TextRenderObject(int capacity) {
		this.allocate(capacity);
	}

	private void allocate(int capacity) {
		capacity = Integer.highestOneBit((capacity - 1) | 1) << 1;
		this.ms = Arena.ofAuto().allocate(TEXT_CHAR_LAYOUT, capacity);
	}

	public void updatePosition(float x, float y) {
		var offX = x - this.x;
		var offY = y - this.y;
		this.x = x;
		this.y = y;
		ms.spliterator(TEXT_CHAR_LAYOUT).forEachRemaining(charMs -> {
			VH_posX.set(charMs, 0L, (float)VH_posX.get(charMs, 0L) + offX);
			VH_posY.set(charMs, 0L, (float)VH_posY.get(charMs, 0L) + offY);
		});
	}

	public void updateTex(int index, byte u, byte v) {
//		Logger.log(() -> FMT."Updating tex coords to \{u}, \{v}");
		var charMs = this.ms.asSlice((long) index * TEXT_CHAR_LAYOUT.byteSize(), TEXT_CHAR_LAYOUT);
		VH_uBase.set(charMs, 0L, u);
		VH_vBase.set(charMs, 0L, v);
	}

	public void updateTex(byte u, byte v) {
//		Logger.log(() -> FMT."Updating tex coords to \{u}, \{v}");
		ms.spliterator(TEXT_CHAR_LAYOUT).forEachRemaining(charMs -> {
			VH_uBase.set(charMs, 0L, u);
			VH_vBase.set(charMs, 0L, v);
		});
	}

	public void updateCharacters(int length, Provider provider) {
		if (length > ms.byteSize() / TEXT_CHAR_LAYOUT.byteSize()) allocate(length);
		var ms = this.ms;

		var xBase = this.x;
		var yBase = this.y;
		var sink = new Sink() {
			private int i;

			@Override
			public void accept(byte character, int color, int flags, float x, float y, float size) {
				int i = this.i;
				if (i >= length) throw new BufferOverflowException("Overflowed specified length");
				VH_posX.set(ms, (long)i, xBase + x);
				VH_posY.set(ms, (long)i, yBase + y);
				VH_char.set(ms, (long)i, character);
//				Logger.log(() -> FMT."Accepted '\{(char)Byte.toUnsignedInt(character)}' (0x%02x\{Byte.toUnsignedInt(character)})");
				VH_color.set(ms, (long)i, color);
				VH_flags.set(ms, (long)i, (byte)flags);
				VH_size.set(ms, (long)i, size);
				this.i = i + 1;
			}
		};
		provider.accept(sink);

		this.length = length;
	}

	@Override
	public int instanceCount() {
		return length;
	}

	@Override
	public int size() {
		return super.size();
	}

	@Override
	public MemoryLayout instanceLayout() {
		return TEXT_CHAR_LAYOUT;
	}

	@Override
	public void uploadTo(dev.pfaff.altargui.render.Sink sink) {
		sink.accept(ms.asSlice(0, TEXT_CHAR_LAYOUT.byteSize() * length));
	}

	@Override
	public void uSpecUploadAtUnchecked(ViaIntermediateBuffer buffer, long offset) {
		super.uSpecUploadAtUnchecked(buffer, offset);
	}

	@Override
	public void uSpecUploadCanvas(CanvasImpl canvas) {
		super.uSpecUploadCanvas(canvas);
	}

	@Override
	public void uSpecUploadLayered(LayeredRenderBuffer buffer) {
		super.uSpecUploadLayered(buffer);
	}

	@Override
	public int program() {
		return textRenderLayer();
	}

	@Override
	public void close() {
	}

	@Override
	public String toString() {
		var sb = new StringBuilder("TextRenderObject[x=");
		sb.append(x);
		sb.append(", y=");
		sb.append(y);
		sb.append(", length=");
		sb.append(length);
		sb.append(", ms=");
		sb.append(ms);
		sb.append(", characters=[");
		var upr = ms.byteSize() / TEXT_CHAR_LAYOUT.byteSize();
		for (int i = 0; i < upr; i++) {
			if (i == 0) {
				sb.append('\'');

			}
			sb.append((char)Byte.toUnsignedInt((byte)VH_char.get(ms, (long)i)));
			if (i == upr - 1) {
				sb.append('\'');
			} else {
				sb.append("', '");
			}
		}
		sb.append("]]");
		return sb.toString();
	}

	@FunctionalInterface
	public interface Provider extends Consumer<Sink> {
		@Override
		void accept(Sink sink);
	}

	@FunctionalInterface
	public interface Sink {
		/**
		 * @param color RGBA
		 * @param flags {@link dev.pfaff.altargui.text.StyleFlags style flags}
		 * @param x the x-offset of the character relative to the render object's position.
		 * @param y the y-offset of the character relative to the render object's position.
		 * @param size the size of the character.
		 */
		void accept(byte character, int color, int flags, float x, float y, float size);
	}
}
