package dev.pfaff.altargui.shader;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.render.GLUtil;
import dev.pfaff.log4truth.Logger;
import dev.pfaff.glass.opengl.enums.GLVertexAttribIType;
import dev.pfaff.glass.opengl.enums.GLVertexAttribLType;
import dev.pfaff.glass.opengl.enums.GLVertexAttribPointerType;
import it.unimi.dsi.fastutil.ints.IntImmutableList;

import java.lang.foreign.MemoryLayout;
import java.lang.foreign.StructLayout;
import java.lang.foreign.ValueLayout;
import java.util.List;
import java.util.Objects;

import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static dev.pfaff.log4truth.StandardTags.WARN;
import static java.lang.foreign.MemoryLayout.sequenceLayout;
import static java.lang.foreign.MemoryLayout.structLayout;
import static java.lang.foreign.ValueLayout.*;

public final class InstanceLayout {
	private static final ValueLayout.OfByte PADDING_LAYOUT = JAVA_BYTE.withName("padding");

	private final List<VertexAttribSpec> attribs;
	/**
	 * The stride between elements in the buffer.
	 */
	private final int stride;
	private final int divisor;
	private final StructLayout layout;

	public InstanceLayout(List<VertexAttribSpec> attribs, int divisor) {
		this.attribs = List.copyOf(attribs);
		this.divisor = divisor;
		this.layout = structLayout(attribs.stream().map(attrib -> {
			if (attrib == VertexAttribSpec.PADDING_BYTE) return PADDING_LAYOUT;
			MemoryLayout layout = attrib.type().layout();
			if (attrib.width() != 1) {
				layout = sequenceLayout(attrib.width(), layout);
			}
			return layout.withName(attrib.name());
		}).toArray(MemoryLayout[]::new));
		this.stride = Math.toIntExact(this.layout.byteSize());
		if (this.layout.byteSize() % 32 != 0) {
			Logger.log(() -> STR."Instance layout is \{this.layout.byteSize()} bytes in size, which is not a multiple of 32 bytes. This may result in suboptimal performance.", List.of(WARN));
		}
	}

	public BuiltInstanceLayout build(ShaderProg shader) {
		GuiGlobals.GL.glUseProgram(shader.handle());
		int offset = 0;
		var builtAttribs = new VertexAttrib[attribs.size()];
		for (int i = 0; i < builtAttribs.length; i++) {
			var attrib = attribs.get(i);
			if (attrib == VertexAttribSpec.PADDING_BYTE) {
				builtAttribs[i] = VertexAttrib.PADDING_BYTE;
				offset += Byte.BYTES;
				continue;
			}
			int location = GLUtil.glGetAttribLocation(shader.handle(), attrib.name());
			if (location == -1)
				throw new IllegalArgumentException(STR."The shader \{shader.name()} does not have the attribute \{attrib.name()}");
			builtAttribs[i] = new VertexAttrib(location, attrib.type(), attrib.normalized(), attrib.width(), offset);
			var i_ = i;
			Logger.log(() -> STR."found attrib \{attrib.name()}: \{builtAttribs[i_]}", List.of(DEBUG));

			offset += (int)layout.select(PathElement.groupElement(i)).byteSize();
		}
		GuiGlobals.GL.glUseProgram(0);
		return new BuiltInstanceLayout(stride, divisor, List.of(builtAttribs));
	}

	public List<VertexAttribSpec> attribs() {
		return attribs;
	}

	public int stride() {
		return stride;
	}

	public int divisor() {
		return divisor;
	}

	public StructLayout layout() {
		return layout;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == this) return true;
		if (obj == null || obj.getClass() != this.getClass()) return false;
		var that = (InstanceLayout) obj;
		return Objects.equals(this.attribs, that.attribs) && this.stride == that.stride && this.divisor == that.divisor;
	}

	@Override
	public int hashCode() {
		int result = 1;
		result = 31 * result + attribs.hashCode();
		result = 31 * result + Integer.hashCode(stride);
		result = 31 * result + Integer.hashCode(divisor);
		return result;
	}

	@Override
	public String toString() {
		return STR."InstanceLayout[attribs=\{attribs}, stride=\{stride}, divisor=\{divisor}]";
	}
}
