package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.widgets.state.StatefulWidgetInstance;
import dev.pfaff.altargui.util.listenable.Signal;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.lang.invoke.VarHandle;
import java.util.function.BiFunction;
import java.util.function.Supplier;

import static java.lang.invoke.MethodHandles.filterCoordinates;
import static java.lang.invoke.MethodHandles.filterValue;

/**
 * A widget that replaces its child using the provided {@code builder} whenever notified by signal.
 */
public record SignalBuilderWidget(Signal signal, Supplier<Widget> builder) implements Widget {
	@Override
	public @NotNull WidgetInstance<SignalBuilderWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	public static final class SignalManager<T, S extends Signal> {
		private final VarHandle VH_signal;
		private final VarHandle VH_listener;
		private final BiFunction<T, S, ListenerDisposer> listenerFactory;

		public SignalManager(MethodHandles.Lookup lookup, String signalField, Class<S> signalType, String listenerField, BiFunction<T, S, ListenerDisposer> listenerFactory) {
			var receiverType = lookup.lookupClass();
			var MH_castReceiver = MethodHandles.explicitCastArguments(MethodHandles.identity(receiverType),
					MethodType.methodType(receiverType, Object.class));
			var MH_castSignalTypeToSignal = MethodHandles.explicitCastArguments(MethodHandles.identity(Signal.class),
					MethodType.methodType(receiverType, signalType));
			var MH_castSignalToSignalType = MethodHandles.explicitCastArguments(MethodHandles.identity(signalType),
					MethodType.methodType(receiverType, Signal.class));
			VarHandle VH_signal, VH_listener;
			try {
				VH_signal = lookup.findVarHandle(receiverType, signalField, signalType);
				VH_listener = lookup.findVarHandle(receiverType, listenerField, ListenerDisposer.class);
			} catch (NoSuchFieldException | IllegalAccessException e) {
				throw new RuntimeException(e);
			}
			this.VH_signal = filterValue(filterCoordinates(VH_signal.withInvokeExactBehavior(), 0, MH_castReceiver),
					MH_castSignalToSignalType,
					MH_castSignalTypeToSignal);
			this.VH_listener = filterCoordinates(VH_listener.withInvokeExactBehavior(), 0, MH_castReceiver);
			this.listenerFactory = listenerFactory;
		}

		private void setNewListener(T receiver, S signal) {
			this.VH_listener.set(receiver, this.listenerFactory.apply(receiver, signal));
		}

		public void configure(T receiver, S newSignal) {
			if ((S) this.VH_signal.get(receiver) != newSignal) {
				var currentListener = (ListenerDisposer) this.VH_listener.get(receiver);
				if (currentListener != null) currentListener.close();
				this.VH_signal.set(receiver, newSignal);
				setNewListener(receiver, newSignal);
			}
		}

		public void onMount(T receiver) {
			if ((ListenerDisposer) this.VH_listener.get(receiver) == null) {
				var signal = (S) this.VH_signal.get(receiver);
				if (signal != null) {
					setNewListener(receiver, signal);
				}
			}
		}

		public void onUnmount(T receiver) {
			var listener = (ListenerDisposer) this.VH_listener.get(receiver);
			if (listener != null) {
				listener.close();
				this.VH_listener.set(receiver, (ListenerDisposer) null);
				this.VH_signal.set(receiver, (Signal) null);
			}
		}
	}

	private static final class State extends StatefulWidgetInstance<SignalBuilderWidget> {
		private static final SignalManager<State, Signal> SIGNAL_MANAGER = new SignalManager<>(MethodHandles.lookup(),
				"signal", Signal.class,
				"listener",
				(state, signal) -> signal.listenAndInvoke(state::markNeedsRebuild)
		);

		private @Nullable Signal signal;
		private Supplier<Widget> builder;

		private @Nullable ListenerDisposer listener;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		protected @NotNull Widget build() {
			return this.builder.get();
		}

		@Override
		public void configure(StateContext stateContext, SignalBuilderWidget widget, boolean init) {
			if (this.builder != widget.builder) {
				this.builder = widget.builder;
				markNeedsRebuild();
			}
			SIGNAL_MANAGER.configure(this, widget.signal);
		}

		@Override
		protected void onMount() {
			super.onMount();

			SIGNAL_MANAGER.onMount(this);
		}

		@Override
		protected void onUnmount() {
			SIGNAL_MANAGER.onUnmount(this);
			super.onUnmount();
		}
	}
}
